
/* ************************************************
   *** This modules test most features of JAVAR ***
   ************************************************ */

class TestAll {

  static final int MAX = 1600;

  // Class variables

  static int[][] a = new int[MAX][MAX];

  // Instance variables

  int[][] b = new int[MAX][MAX];

  double[] d = new double[MAX];

  // Constructor with trivial DOACROSS-loop

  TestAll() {
    d[0] = 1.0;
    /*par cyclic, sync=1 */
    for (int i = 1; i < MAX; i++) {
      /*wait 0, i-1 */
      d[i] = d[i-1];
      /*post 0 */
    }
  }

  public static void main(String arg[]) {
    long    t1, t2;

    // Test 1

    t1 = System.currentTimeMillis();
    TestAll o = new TestAll();
    t2 = System.currentTimeMillis();

    System.out.println("time TestAll() = " + (t2 - t1) / 1000.0);

    for (int i = 0; i < MAX; i++)
      if (o.d[i] != 1.0) {
        System.out.println("ERROR 1");
        System.exit(1);
      }

    // Test 2

    for (int k = 0; k <= MAX; k+=400) {

      for (int i = 0; i < MAX; i++)
        for (int j = 0; j < MAX; j++)
          a[i][j] = 2;

      t1 = System.currentTimeMillis();
      classMethod(k);
      t2 = System.currentTimeMillis();

      System.out.println("time classMethod(" + k + ") = " + (t2 - t1) / 1000.0);

      for (int i = 0; i < k; i++)
        for (int j = 0; j < k; j++)
  	  if (a[i][j] != -2) {
	    System.out.println("ERROR 2");
	    System.exit(1);
          }
      for (int i = k; i < MAX; i++)
        for (int j = k; j < MAX; j++)
  	  if (a[i][j] != 2) {
	    System.out.println("ERROR 2");
	    System.exit(1);
          }
    }

    // Test 3

    for (int k = 0; k <= MAX; k+=400) {

      for (int i = 0; i < MAX; i++)
        for (int j = 0; j < MAX; j++)
          o.b[i][j] = 3;

      t1 = System.currentTimeMillis();
      o.instanceMethod(k);
      t2 = System.currentTimeMillis();

      System.out.println("time instanceMethod(" + k + ") = " + (t2 - t1) / 1000.0);

      for (int i = 0; i < k; i++)
        for (int j = 0; j < k; j++)
  	  if (o.b[i][j] != -3) {
	    System.out.println("ERROR 3");
	    System.exit(1);
          }
      for (int i = k; i < MAX; i++)
        for (int j = k; j < MAX; j++)
  	  if (o.b[i][j] != 3) {
	    System.out.println("ERROR");
	    System.exit(1);
          }
    }

    // Test 4

    for (int k = 0; k <= MAX; k+=400) {

      for (int i = 0; i < MAX; i++)
        for (int j = 0; j < MAX; j++)
          o.b[i][j] = 4;

      t1 = System.currentTimeMillis();
      o.instanceMethod_stride(k);
      t2 = System.currentTimeMillis();

      System.out.println("time instanceMethod_stride(" + k + ") = " + (t2 - t1) / 1000.0);

      for (int i = 10; i <= k/2; i+=2)
        for (int j = 0; j < k; j++)
  	  if (o.b[i][j] != -4) {
	    System.out.println("ERROR 4");
	    System.exit(1);
          }
	  else
	   o.b[i][j] = 4;
      for (int i = 0; i < MAX; i++)
        for (int j = 0; j < MAX; j++)
  	  if (o.b[i][j] != 4) {
	    System.out.println("ERROR 4");
	    System.exit(1);
          }
    }

    // Test 5

    for (int k = 0; k <= MAX; k+=400) {

      for (int i = 0; i < MAX; i++)
        for (int j = 0; j < MAX; j++)
          o.b[i][j] = 5;

      t1 = System.currentTimeMillis();
      o.instanceMethod_bnd(k);
      t2 = System.currentTimeMillis();

      System.out.println("time instanceMethod_bnd(" + k + ") = " + (t2 - t1) / 1000.0);

      for (int i = k-1; i > 17; i-=6)
        for (int j = 0; j < k; j++)
  	  if (o.b[i][j] != -5) {
	    System.out.println("ERROR 5");
	    System.exit(1);
          }
	  else 
	   o.b[i][j] = 5;
      for (int i = 1; i < k; i++)
        for (int j = 0; j < k; j++)
  	  if (o.b[i][j] != 5) {
	    System.out.println("ERROR 5");
	    System.exit(1);
          }
    }
    // Test 6

    System.out.println("\nThe following loops should print the integers 1..10");

    /*par */
    for (int i = 1; i <= 10; i++)
      System.out.print(" " +  i );
    System.out.println(" <<");
    /*par */
    for (int i = 0; i < 10; i++)
      System.out.print(" " +  (i+1) );
    System.out.println(" <<");
    /*par */
    for (int i = 10; i >= 1; i--)
      System.out.print(" " + i );
    System.out.println(" <<");
    /*par */
    for (int i = 10; i > 0; i--)
      System.out.print(" " + i );
    System.out.println(" <<");
    /*par */
    for (int i = 2; i <= 20; i += 2)
      System.out.print(" " + (i/2) );
    System.out.println(" <<");
    /*par */
    for (int i = 2; i < 21; i += 2)
      System.out.print(" " + ( i/2) );
    System.out.println(" <<");
    /*par */
    for (int i = 20; i >= 2; i -= 2)
      System.out.print(" " + ( i/2 ));
    System.out.println(" <<");
    /*par */
    for (int i = 20; i > 1; i -= 2)
      System.out.print(" " + ( i/2 ));
    System.out.println(" <<");
    /*par */
    for (int i = 3; i <= 30; i += 3)
      System.out.print(" " + ( i/3) );
    System.out.println(" <<");
    /*par */
    for (int i = 3; i <= 31; i += 3)
      System.out.print(" " + ( i/3) );
    System.out.println(" <<");
    /*par */
    for (int i = 3; i <= 32; i += 3)
      System.out.print(" " + ( i/3) );
    System.out.println(" <<");
    /*par */
    for (int i = 3; i < 31; i += 3)
      System.out.print(" " + ( i/3 ) );
    System.out.println(" <<");
    /*par */
    for (int i = 3; i < 32; i += 3)
      System.out.print(" " + ( i/3 ) );
    System.out.println(" <<");
    /*par */
    for (int i = 3; i < 33; i += 3)
      System.out.print(" " + ( i/3 ) ) ;
    System.out.println(" <<");
    /*par */
    for (int i = 9; i < 99; i += 9)
      System.out.print(" " + (  i/9 ) );
    System.out.println(" <<");
    /*par */
    for (int i = 11; i <= 93; i += 9)
      System.out.print(" " + (  (i-2)/9 ) );
    System.out.println(" <<");
    /*par */
    for (int i = 11; i < 96; i += 9)
      System.out.print(" " + (  (i-2)/9 ) );
    System.out.println(" <<");
    int local = -10;
    /*par */
    for (int i = (5+local*2)+17; i < 17+(-local)*2-16; i += 2)
      System.out.print(" " + (  i/2 ) );
    System.out.println(" <<");

    System.out.println("\nThe following three test3 should transfer exceptions back");
    System.out.println("(but parallel versions may become undeterministic)\n");

    // Test 7

    for (int k = 0; k < 4; k++) {

    for (int i = 0; i < MAX; i++)
      a[i][0] = 7;

    try {
      classMethod_exc();
    }
    catch (Exception e) { 
      int t = 0;
      for (int i = 0; i < MAX; i++) 
	if (a[i][0] == -7) t++;
      System.out.println("Exception correctly transferred back from loop, "  
	+ t + " iterations done");
    }
    catch (Throwable e) { 
      int t = 0;
      for (int i = 0; i < MAX; i++) 
	if (a[i][0] == -7) t++;
      System.out.println("Throwable correctly transferred back from loop, "  
	+ t + " iterations done");
    }
    }

    // Test 8

    for (int k = 0; k < 4; k++) {

    for (int i = 0; i < MAX; i++)
      a[i][0] = 8;

    try {
      traverse1(0, MAX-1);
    }
    catch (Exception e) { 
      int t = 0;
      for (int i = 0; i < MAX; i++) 
	if (a[i][0] == -8) t++;
      System.out.println("Exception correctly transferred back from traverse1, "  
	+ t + " elements set");
    }
    catch (Throwable e) { 
      int t = 0;
      for (int i = 0; i < MAX; i++) 
	if (a[i][0] == -8) t++;
      System.out.println("Throwable correctly transferred back from traverse1, "  
	+ t + " elements set");
    }
    }

    // Test 9 

    for (int k = 0; k < 4; k++) {

    for (int i = 0; i < MAX; i++)
      a[i][0] = 9;

    try {
      traverse2(0, MAX-1);
    }
    catch (Exception e) { 
      int t = 0;
      for (int i = 0; i < MAX; i++) 
	if (a[i][0] == -9) t++;
      System.out.println("Exception correctly transferred back from traverse2, "  
	+ t + " elements set");
    }
    }

    // Test 10 

    for (int k = 0; k < MAX; k+=400) {

      for (int i = 0; i < k; i++) 
  	 a[0][i] = k-i-1;
      for (int i = k; i < MAX; i++) 
  	 a[0][i] = 10;

       t1 = System.currentTimeMillis();
       quicksort(a[0], 0, k-1);
       t2 = System.currentTimeMillis();

      System.out.println("time quicksort(" + k + ") = " + (t2 - t1) / 1000.0);

      for (int i = 0; i < k; i++)
         if  (a[0][i] != i) {
  	   System.out.println("ERROR 10");
	   System.exit(1);
       } 
      for (int i = k; i < MAX; i++)
         if  (a[0][i] != 10) {
  	   System.out.println("ERROR 10");
	   System.exit(1);
       } 

    }

    // Test 11

   Tree2 root = new Tree2(1.0);
    for (int k = 0; k < 2000; k++)
	 root.insert( new Tree2(400.0) );
    System.out.println("seeing height " + Tree2.compHeight1(root));
    root.test(11);
    System.out.println("seeing height " + root.compHeight2());
    root.test(-11);

  }

  static void classMethod(int k) {
    /*par block, multi=400 */
    for (int i = 0; i < k; i++)
      for (int j = 0; j < k; j++)
        a[i][j] = -2;
  }
  void instanceMethod(int k) {
    /*par guided, multi=600 */
    for (int i = 0; i < k; i++)
      for (int j = 0; j < k; j++)
        b[i][j] = -3;
  }
  void instanceMethod_stride(int k) {
    /*par  */
    for (int i = 10; i <= k/2; i+=2)
      for (int j = 0; j < k; j++)
        b[i][j] = -4;
  }
  void instanceMethod_bnd(int k) {
    /*par threads=16 */
    for (int i = k-1; i > 17; i-=6)
      for (int j = 0; j < k; j++)
        b[i][j] = -5;
  }
  static void classMethod_exc() throws Throwable {
    /*par block, (Exception) , (Throwable)  */
    for (int i = 0; i < MAX; i++) {
      if (i == 500)
	throw new Exception();
      else if (i == 800)
	throw new Throwable();
      else
          a[i][0] = -7;
    }
  }

  final static void traverse1(int low, int high) throws Throwable {

    if (low > high) 
     return;
    else {

       int mid = (low+high)/2;

      if (mid == 500)
       throw new Exception();
      else if (mid == 800)
       throw new Throwable();

       a[mid][0] = -8;

      /*par_inv cut_depth=3, no_overhead, (Exception), (Throwable) */
      {
        traverse1(low,mid-1);
        traverse1(mid+1, high);
       }
    }
  }

  final static void traverse2(int low, int high) throws Exception {
    if (low > high) 
     return;
    else {
       int mid = (low+high)/2;

      if (mid == 500)
       throw new Exception();

       a[mid][0] = -9;

      /*par_inv cut_depth=6, overhead, (Exception) */
      {
        traverse2(low,mid-1);
        traverse2(mid+1, high);
       }
    }
  }
  private static void swap(int [] a, int i, int j) {
    int t = a[i]; a[i] = a[j]; a[j] = t; 
  }

  final static void quicksort(int[] a, int q, int r) {
    if (q < r) { 
      swap(a, q, (q+r)/2); 
      int s = q;
      for (int i = q+1; i <= r; i++)
        if (a[i] <= a[q])  
	  swap(a, ++s, i);
      swap(a, q, s);
      /*par_inv*/
      {
        quicksort(a, q,   s-1);
        quicksort(a, s+1, r  );
      }
    }
  }
}


class Tree2 {
  double val;
  Tree2   left;
  Tree2   right;

  Tree2(double newval) {
    val   = newval;
    left  = null;
    right = null;
  }

  // Class Method for Computing Height

  final static int compHeight1(Tree2 t) {
    if (t == null)
      return 0;
    else {
      int l, r;

      t.val = 11;
 
      /*par_inv */ 
      {
        l = compHeight1(t.left);
        r = compHeight1(t.right);
      }     
      return (l > r) ? (l+1) : (r+1);
    }
  }

  // Instance Method for Computing Height
  // (expressed in rigid form required by JAVAR)

   final int compHeight2() {
    int l = 0 , r = 0;

    val = -11;

    if ((left != null) && (right != null))
    /*par_inv */
    {
      l = left.compHeight2();
      r = right.compHeight2();
    }
    else if  (left != null)
      l = left.compHeight2();
    else if  (left != null)
      r = right.compHeight2();

    return (l > r) ? (l+1) : (r+1);
  }

  // Auxiliary Instance Methods

  void test(double d) {
    if (val != d) {
      System.out.println("ERROR " + d);
      System.exit(1);
    }
    if (left != null)
      left.test(d);
    if (right != null)
      right.test(d);
  }

  void insert(Tree2 node) {
    if (left == null)
      left = node;
    else if (right == null)
      right = node;
    else if (val >= 0) {
       val = -400.0d;
       left.insert(node);
    }
    else {
      val = 400.0d;
      right.insert(node);
    }
  }
}
