/*  **********************************
    *** LoopPool of Iterations     ***
    *** (C) 1996/1997 Aart Bik     ***
    *** Indiana University         ***
    ********************************** */

package parlooppack;

public class LoopPool implements Schedules {

  // Instance Variables

  private int low, high, sched, numW, blocksz, count;

  // Constructor

  public LoopPool(int l, int h, int n, int s) {
    low   = l;
    high  = h;
    numW  = n; 
    sched = s;

    blocksz = (int) Math.ceil(((double) (high-low)) / numW);
    count   = numW;
  }

  // Instance Method

  public final synchronized boolean nextWork(LoopWorker worker) {
    boolean more = false;

    switch (sched) {

      case SCHED_CYCLIC:
	more          = (count-- > 0);
        worker.low    = low++;
        worker.high   = high;
        worker.stride = numW;
	break;

      case SCHED_GUIDED:
        blocksz = (int) Math.ceil(((double) (high-low)) / numW);

      case SCHED_BLOCK:                  // Fall Through 
	more          = (low < high);
        worker.low    = low;
        worker.high   = Math.min(low + blocksz, high);
        worker.stride = 1;
	low          += blocksz;
	break;
    }
    return more;
  }
}

