/*  **********************************
    *** Random Synchronization     ***
    *** (C) 1996/1997 Aart Bik     ***
    *** Indiana University         ***
    ********************************** */

package parlooppack;

public class RandomSync {

  // Instance Variables

  private          int     low;
  private volatile boolean bits[]; // NOTE: here in fact the ELEMENTS
                                   // of bits should be declared volatile

  public RandomSync(int l, int h) {
    if (l < h)
      bits = new boolean[h-l];
    low = l;
  }

  // Instance Methods

  // Default suspended-waiting implemenation of post/wait

  public final synchronized void doPost(int i) {
    bits[i-low] = true;
    notifyAll(); 
  }

  public final synchronized void doWait(int i) {
    if (low <= i) 
      while (! bits[i-low]) { 
        try { wait(); } 
        catch (InterruptedException e) { }
      }
  }

  // A potentially more efficient but 'dangerous' 
  // busy-waiting implementation of post/wait
  //
  // Dangerous because:
  //
  // 1) Java does not garantuee fairness
  // 2) volatile does NOT work on the elements of an array 
  //    so that changes to the array bits may remain 
  //    invisible to other threads...

  public final void doPost_bw(int i) {
    bits[i-low] = true;
  }

  public final void doWait_bw(int i) {
    if (low <= i) 
      while (! bits[i-low]) 
        Thread.currentThread().yield();
  }
}
