

/* *************
   *** JAVAR ***
   ****************************************************
   *** Copyright (c) 1996/1997                      ***
   *** Aart J.C. Bik             Indiana University *** 
   *** All Rights Reserved                          ***
   ****************************************************
   *** Please refer to the LICENSE file distributed ***
   *** with this software for further details on    ***
   *** the licensing terms and conditions.          ***
   ***                                              ***
   *** Please, report all bugs, comments, etc.      ***
   *** to: ajcbik@extreme.indiana.edu               ***
   ****************************************************
   *** parallel.c : make implicit parallelism explicit
   ***              by means of JAVA multi-threading
   ***
   ***
   *** Your courtesy in mentioning the use of this compiler 
   *** in any scientific work that presents results obtained
   *** by using (extensions or modifications of) this compiler
   *** is highly appreciated.
   ***
   *** NOTE! 
   *** This implementation does NOT provide a complete Java 
   *** front-end (unicode escapes are not supported
   *** and the semantic analysis is rather limited).
   ***
   *** */

/* ********************************************************
   *** INCLUDE FILES and DEFINITIONS                    ***
   ********************************************************  */

#include "global.h"
#include "y.tab.h"

#define UNSAFE(e) (!((!e)||(e->kind!=ID_)||(! in_list(e))))
#define MK_LIT(l) make_literal(INT_T, NULL, (l))

/* ********************************************************
   *** EXTERNAL VARIABLES                               ***
   ******************************************************** */

/* PRIVATE
   ******* */

/* general information */

static int      par_cnt; 
static stmt_ptr my_class, my_method, my_annot;

/* loop information */

static stmt_ptr my_loop;
static expr_ptr my_name,  my_init, my_bound;
static symb_ptr my_sync;
static short    my_strict; 
static long     my_stride;

/* lists */

static int       ipos;
static short    *imark = NULL;
static expr_ptr *elist = NULL;
static stmt_ptr *slist = NULL;

/* flags */

static short    my_flag;  
static expr_ptr my_global;
static int      self_inv;

/* hooks */

static stmt_ptr f_hook, c_hook, i_hook;           /* used for Exceptions */
static symb_ptr t_hook, d_hook, p_hook, s_hook;

/* *********************************************************
   *** PRIVATE FUNCTIONS                                 ***
   ********************************************************* */

/* Test Equivalences of Names
   ************************** */

static short is_name(expr_ptr n1, expr_ptr n2) {
   return ( (n1) && (n1 -> kind == ID_) &&                          
            (n1 -> u.name.entry == n2 -> u.name.entry) &&  
            (n1 -> u.name.scope == n2 -> u.name.scope) &&  
            (n1 -> u.name.info  == n2 -> u.name.info)   );
}

/* List Routines
   ************* */

static void mark_list(expr_ptr e) {
  int k;
  for (k = 0; k < ipos; k++)
    if (is_name(e, elist[k])) {
      imark[k] = 1;
#ifdef DEBUG_LIST
  fprintf(stderr, "LIST : mark %s\n", id_table(e -> u.name.entry));
#endif
      break;
    }
}

static short in_list(expr_ptr e) {
  int k;
  for (k = 0; k < ipos; k++)
    if (is_name(e, elist[k]))
      return 1;        
  return 0;
}

static void add_list(expr_ptr e, stmt_ptr s) {

  static isz = 0;

  /* Search List */

  if (in_list(e))
    return;

  /* New ID */

  if (ipos >= isz)  {
    isz += 200;
    if (imark) {
      imark = (short    *) more_mem(imark, isz * sizeof(short)); 
      elist = (expr_ptr *) more_mem(elist, isz * sizeof(expr_ptr)); 
      slist = (stmt_ptr *) more_mem(slist, isz * sizeof(stmt_ptr)); 
    }
    else {
      imark = (short    *) make_mem(isz * sizeof(short)); 
      elist = (expr_ptr *) make_mem(isz * sizeof(expr_ptr)); 
      slist = (stmt_ptr *) make_mem(isz * sizeof(stmt_ptr)); 
    }
  }

  if  ( ((e) && (e -> kind != ID_)) ||
        ((s) && (! s -> u.decl.name))  )  /* Safety */
    javar_out(-1, "corrupt AST in add_list()");

#ifdef DEBUG_LIST
  fprintf(stderr, "LIST : add  %s\n", id_table(e -> u.name.entry));
#endif
  
  imark[ipos  ] = 0;
  elist[ipos  ] = e;
  slist[ipos++] = s;
}

/* Strip Annotations of serial-branch (TRAVERSAL ROUTINE)
   ****************************************************** */

static void strip_annot(stmt_ptr s) {
  if (s -> kind == ANNOT_S) {

    /* Convert into Dummy Concatenation Nodes */

    s -> kind        = CONCAT_S;
    s -> u.concat.s1 = s -> u.annot.body;
    s -> u.concat.s2 = NULL;
  }
}

/* *********************************************************
   *** PRIVATE FUNCTIONS RELATED TO LOOP PARALLELIZATION ***
   ********************************************************* */

/* Check loop-invariance of variables in list
   ****************************************** */

/* Collection of VARIABLE Names (TRAVERSAL ROUTINE) */

static void collect_names(expr_ptr e) {
  if (my_flag)
    return;
  else if (e -> kind == 'i') {
    my_flag = 1;
    javar_out(0, "parallelization prohibited by method invocation in loop bound");
  }
  else if (e -> kind == '.') {
    my_flag = 1;
    javar_out(0, "parallelization prohibited by field access in loop bound");
  }
  else if (e -> kind == '[') {
    my_flag = 1;
    javar_out(0, "parallelization prohibited by array element access in loop bound");
  }
  else if ((e -> kind == ID_) && (e -> u.name.lcomp)) {
    my_flag = 1;
    javar_out(0, "parallelization prohibited by qualified name in loop bound");
  }
  else if ( (e -> kind == ID_)            &&
            (e -> u.name.info != METH_N)  &&
            (e -> u.name.info != PACK_N)  &&
            (e -> u.name.info != TYPE_N)  &&
            (e -> u.name.info != LABEL_N)  ) {
    add_list(e, NULL);
    if (e -> u.name.info != LOCPAR_N)  {
      my_global = e;
    }
    else if (is_name(e, my_name)) {
      my_flag = 1;
      javar_out(0, "parallelization prohibited by complex loop bound");
    }
  }
}

/* Collection of Local/Parameter Declarations  */

static void collect_decl(stmt_ptr s) {
  if (s -> kind == DECL_S) 
    add_list(s -> u.decl.name, s);
}

/* Test for Occurrence of break/continue/return (TRAVERSAL ROUTINE) */

static void no_transfer(stmt_ptr s) {
  if (my_flag)
    return;
  else if ((s -> kind == BRK_S) || (s -> kind == CONT_S) || (s -> kind == RET_S)) {
    javar_out(0, "parallelization prohibited by break/continue/return in body");
    my_flag = 1;
  }
}

/* Test LHS-Occurrence of a Name in List and mark used Names (TRAVERSAL ROUTINE) */

static void mark_and_invlhs(expr_ptr e) {
  if (my_flag)
    return;
  else if ((my_global) && ((e -> kind == '.') ||  
			  ((e -> kind == ID_) && (e -> u.name.lcomp)))) {
    my_flag = 1; 
    javar_out(0, "parallelization prohibited by potential change of %s in field access",
			  id_table(my_global -> u.name.entry));
  }
  else if (e -> kind == ID_)  
    mark_list(e); 
  else if ((e -> kind == 'i') && (my_global)) {  /* VERY conservative test */
    expr_ptr nm = e -> u.op.lhs;
    my_flag = 1; 
    javar_out(0, "parallelization prohibited by potential change of %s in %s()",
			  id_table(my_global -> u.name.entry),
			  (nm) ? id_table(nm -> u.name.entry) : "?");
  }
  else if ( (((e -> kind == '=') || ((e -> kind >= PIS_) && (e -> kind <= RRRIS_)) ) &&
             (UNSAFE(e -> u.op.lhs)))  ||
            (((e -> kind == PP_)     || (e -> kind == MM_)) &&
	    ((UNSAFE(e -> u.op.lhs)) || (UNSAFE(e -> u.op.rhs)))) ) { 
    my_flag = 1;
    javar_out(0, "parallelization prohibited by assignment to %s",
	        id_table(e -> u.op.lhs -> u.name.entry));
  }
}

/* Check loop-invariance of loop bounds
   ************************************ */ 

static short bounds_not_inv(void) {

  my_global = NULL;
  ipos      = 0;
  my_flag   = 0;

  /* Collect IDs in Bounds */

  expr_traverse(my_init,  NULL, collect_names);
  expr_traverse(my_bound, NULL, collect_names);

  if ((! my_flag)  && (ipos > 0))
    stmt_traverse(my_loop, NULL, mark_and_invlhs); 

  return my_flag;
}

/* Check loop-invariance of Locals/Parameters
   ****************************************** */

static short locpars_not_inv(void) {
  ipos    = 0;
  my_flag = 0;

  /* Collect Locals/Parameters of myMethod() */

  stmt_traverse(my_method, collect_decl, NULL); 
  stmt_traverse(my_loop -> u.control.s1, no_transfer, 
					 (ipos > 0) ? mark_and_invlhs : NULL);

  /* Traverse ALSO my_init in case loop normalization will be applied! */

  if ((my_stride != 1) && (ipos > 0))
    expr_traverse(my_init, NULL, mark_and_invlhs);

  return my_flag;
}

/* Hook (unique) Local/Parameter Declarations to Decl. List 
   ******************************************************** */   

static expr_ptr hook_args(expr_ptr args, short formal, short make_unique) {
  int      i;
  expr_ptr d;

  for (i = 0; i < ipos; i++) 
    if (imark[i]) {

      if ((formal) && (make_unique))
	javar_out(-1, "make first actuals unique in hook_args()");
      if (formal) 
        d  = make_declexpr( 
	      give_m_t_c(0, slist[i] -> u.decl.tp, 
	      make_decl(elist[i] -> u.name.entry, 0, NULL), LOCPAR_N) ); 
      else if (make_unique) 
        d = elist[i] = make_name(NULL, 
			unique_table( id_table( elist[i] -> u.name.entry ) ),
			AMBI_N);
      else
	d = make_name(NULL, elist[i] -> u.name.entry, AMBI_N);

      if (args)
	args = make_op(',', args, d);
      else
	args = d;
    }
  return args;
}

static stmt_ptr set_args(stmt_ptr list, short set) {
  int      i;
  stmt_ptr ass;

  for (i = 0; i < ipos; i++) 
    if (imark[i]) {
      
      if (set) 
	ass  = make_simple(EXP_S, make_op('=', 
	 make_op('.',  make_thissup(THIS_), 
	     make_name(NULL, elist[i] -> u.name.entry, MISCF_N)), 
	     make_name(NULL, elist[i] -> u.name.entry, LOCPAR_N)) );
      else
	ass  = give_m_t_c(0, slist[i] -> u.decl.tp, 
		 make_decl(elist[i] -> u.name.entry, 0, NULL), FIELD_N); 

      list = concat(list, ass);
    }
  return list;
}

/* Add Exception Fields to Class (and replace by names)
   + f_hook : field declaration list
   + c_hook : catch-clause list
   + u_hook : re-throw statement list
   **************************************************** */

static void hook_exceptions(expr_ptr nm, symb_ptr i_x, symb_ptr wrk_x) {
  if (nm) {
    if (nm -> kind == ',') {
      hook_exceptions(nm -> u.op.lhs, i_x, wrk_x); 
      hook_exceptions(nm -> u.op.rhs, i_x, wrk_x); 
    }
    else {

      /* Field Declaration: Exception e_x */

      symb_ptr e_x = unique_table("e");
      stmt_ptr f   = give_m_t_c(0, make_c_i_type(copy_expr(nm)), 
		        make_decl(e_x, 0, NULL), FIELD_N);

      /* Catch Clause: catch(Exception e_x) { this.e_x = e_x; } */

      expr_ptr 
	  decl    = make_declexpr(give_m_t_c(0, make_c_i_type(copy_expr(nm)),
                    make_decl(e_x, 0, NULL), LOCPAR_N)); 
      stmt_ptr 
           assign = make_simple(EXP_S, make_op('=', 
                        make_op('.',  make_thissup(THIS_), 
                        make_name(NULL, e_x, MISCF_N)), 
                        make_name(NULL, e_x, LOCPAR_N)) ),
           c      = make_header(CATCH_S, 0, NULL, NULL, 0, decl, NULL, assign);

      f_hook = concat(f_hook, f); 
      c_hook = concat(c_hook, c); 

      /* Re-Throw Clause: 
	  + if (worker_x[i_x].e_x != null) throw worker_x[i_x].e_x; 
	  + if (wi_x.e_x          != null) throw wi_x.ex;
         ********************************************************** */

      if (wrk_x) {
	expr_ptr we = make_op('.', make_op('[',
		       make_name(NULL, wrk_x, LOCPAR_N), 
		       make_name(NULL, i_x,   LOCPAR_N)), 
		       make_name(NULL, e_x,   MISCF_N)),
                comp = make_op(NE_, we, make_literal(-2, "null", 0));

	stmt_ptr
	  t = make_simple(THROW_S, copy_expr(we)), 
	  i = make_control1(IF_S, comp, t); 

	i_hook = concat(i_hook, i);
      }
      else {
	expr_ptr we = make_op('.', 
		       make_name(NULL, i_x,   LOCPAR_N), 
		       make_name(NULL, e_x,   MISCF_N)),
                comp = make_op(NE_, we, make_literal(-2, "null", 0));

	stmt_ptr
	  t = make_simple(THROW_S, copy_expr(we)), 
	  i = make_control1(IF_S, comp, t); 

        i_hook = concat(i_hook, i);
      }
    }
  }
}

/* Add a Try/Catch Class for Exception Fields
   ****************************************** */

static stmt_ptr add_try(stmt_ptr s) {
  return (c_hook)
	 ?  make_control(TRY_S, NULL, make_block(s), c_hook)
	 : s;
}

/* Random Synchroniation Generation (TRAVERSAL ROUTINE)
   **************************************************** */

static void rand_sync(stmt_ptr s) {

  if (s -> kind == ANNOT_S) {
    
    if (! s -> u.annot.body) {
    
      long k = s -> u.annot.f1;
    
      if ((k < 0) || (k >= my_annot -> u.annot.f2)) {
        javar_out(1, "undefined synchronization variable ignored");
        strip_annot(s);
      }
      else {

        symb_ptr pw  = 
	  (my_annot -> u.annot.f5) 
	  ? insert_table( (s -> u.annot.arg) ? "doWait_bw" : "doPost_bw")
	  : insert_table( (s -> u.annot.arg) ? "doWait"    : "doPost");
        expr_ptr arg = (s -> u.annot.arg) 
                     ? copy_expr(s -> u.annot.arg) 
                     : make_name(NULL, my_name -> u.name.entry, AMBI_N);
        expr_ptr inv = make_op('.', make_op('[', 
  		      make_name(NULL, my_sync, AMBI_N), MK_LIT(k)),
                        make_name(NULL, pw, MISCM_N));

        s -> kind  = EXP_S;
        s -> u.arg = make_op('i', inv, arg);
      }
    }
    else {
      javar_out(1, "parallel annotation within parallel loop ignored");
      strip_annot(s);
    }
  }
}

/* Loop Index Normalization
   ************************ */

static stmt_ptr norm_stmt(void) {
  long s       = (my_stride > 0) ? my_stride : (- my_stride);
  expr_ptr off = 
    (s == 1) 
    ? make_name(NULL, my_name -> u.name.entry, LOCPAR_N)
    : make_op('*', make_name(NULL, my_name -> u.name.entry, LOCPAR_N), MK_LIT(s));
      
  /* A Small Hard-Coded Expression Optimizer
     (TYPE is already checked to be integer)  */

  long     opt  = ((my_init -> kind == LITERAL_) &&
		   (my_init -> u.lit.set)        &&
		   (my_init -> u.lit.val.l == 0));
  expr_ptr base = (opt) ? NULL : copy_expr(my_init);
  expr_ptr new  = NULL;
   
  if (my_stride > 0) {
    if (opt) 
      new = off;
    else 
      new = make_op('+', base, off);
  }
  else 
    new = make_op('-', base, off);

  return give_m_t_c(0, make_primitive_type(INT_T), 
	  make_decl(d_hook, 0, new), LOCPAR_N);
}

/* Loop Index Replacement after Normalization (TRAVERSAL ROUTINE)
   ************************************************************** */

static void norm_index(expr_ptr e) {
  if (is_name(e, my_name)) {

    /* OVERWRITE CURRENT EXPRESSION */

    e -> u.name.id    = 
    e -> u.name.entry = d_hook;
    e -> u.name.lcomp = NULL;
    e -> u.name.scope = -1;
    e -> u.name.info  = LOCPAR_N;
  }
}

/* STEP 1: Replace Original Loop with Parallel Loop Construct
   ********************************************************** */

static stmt_ptr replace_loop(void) {
  
  /* Construct new IDs */ 

  symb_ptr
    i_x   = unique_table("i"),
    wrk_x = unique_table("worker"),
    ploop = insert_table("parlooppack.LoopWorker.parloop");

  /* Construct Declaration */

  short is_static = my_method -> u.header.mod & STA_B;

  expr_ptr 
    init = make_te(NEW_, make_c_i_type(make_name(NULL, t_hook, TYPE_N)), 
       make_op(']', NULL, make_op('[', NULL, MK_LIT(my_annot -> u.annot.f1))));
  stmt_ptr  
    ds = give_m_t_c(0, make_c_i_type(make_name(NULL, t_hook, TYPE_N)),
	                    make_decl(wrk_x, 1, init), LOCPAR_N); 

  /* FOR-loop */

  expr_ptr 
    cexpr = make_op(';', make_declexpr(give_m_t_c(0, 
	      make_primitive_type(INT_T), make_decl(i_x, 0, 
	      MK_LIT(0)), LOCPAR_N)), 
              make_op(';', make_op('<', make_name(NULL, i_x, LOCPAR_N), 
	      MK_LIT(my_annot -> u.annot.f1)),
              make_op(PP_, make_name(NULL, i_x, LOCPAR_N), NULL))),
    cargs = (is_static)  ? NULL : make_thissup(THIS_);

  stmt_ptr 
    bd   = make_simple(EXP_S, make_op('=', 
            make_op('[', make_name(NULL, wrk_x, LOCPAR_N), 
			 make_name(NULL, i_x,   LOCPAR_N)),
	    make_te(NEW_, make_c_i_type(
			 make_name(NULL, t_hook, TYPE_N)), hook_args(cargs, 0, 0))));

  /* Construct parloop invocation */

  int      opt1, opt2;
  expr_ptr new_low, new_up, args;
  stmt_ptr loop,    inv,    block;

  loop = make_control1(FOR_S, cexpr, bd);  

  /* Obtain Fresh Copies */

  new_low = copy_expr(my_init);
  new_up  = copy_expr(my_bound);

  /* A Small Hard-Coded Expression Optimizer
     (type is already checked to be integer) */

  opt1 = ((new_low -> kind == LITERAL_) &&
	  (new_low -> u.lit.set)        &&
	  (new_low -> u.lit.val.l == 0) );
  opt2 = ((new_up  -> kind == LITERAL_) &&
	  (new_up  -> u.lit.set)        &&
          (new_up  -> u.lit.val.l == 0) );

  if (my_stride == 1) {
     if (! my_strict)
       new_up = make_op('+', new_up, MK_LIT(1));
  }
  else {
    long s = (my_stride > 0) ? my_stride : (- my_stride);

    if ((my_stride > 0) && (! opt1))
      new_up =  make_op('-', new_up, new_low);
    else if (my_stride < 0)  
      new_up = (opt2) ? new_low : make_op('-', new_low, new_up);

    if (my_strict)
      new_up = make_op('-', new_up, MK_LIT(1));

    /* Add |Stride| */

    new_up = make_op('+', MK_LIT(s), new_up);

    if (s != 1)
      new_up  = make_op('/', new_up, MK_LIT(s));

    new_low = MK_LIT(0); opt1 = 1;
  }

  args = make_op(',', new_low, make_op(',', new_up,
	 make_op(',', make_name(NULL, wrk_x, LOCPAR_N), 
	 make_op(',', MK_LIT(my_annot -> u.annot.f2), 
	    MK_LIT(my_annot -> u.annot.f4)))));
  inv  = make_simple(EXP_S, make_op('i', make_name(NULL, ploop, METH_N), args));

  /* Make Exception Lists */

  f_hook = c_hook = i_hook = NULL;

  hook_exceptions(my_annot -> u.annot.arg, i_x, wrk_x);

  if (i_hook) {
   stmt_ptr loop2 = make_control1(FOR_S, copy_expr(cexpr), make_block(i_hook));  
   inv = concat(inv, loop2); 
  }

  /* Construct Block */

  block = make_block( concat(ds, concat(loop, inv)) );

  if (my_annot -> u.annot.f3) {

    /* Multi-version code */

    expr_ptr comp = make_op(LE_, 
     (opt1) ? copy_expr(new_up) 
            : make_op('-', copy_expr(new_up), copy_expr(new_low)),
	              MK_LIT(my_annot -> u.annot.f3));

    block = make_control(IF_S, comp, my_loop, block);
  }

  /* Return new Construct */

  return block;
}

/* STEP 2: Construction of new run_x()-method
   ****************************************** */

static symb_ptr add_run(char *run_name) {

  /* Construct new IDs */   

  symb_ptr
    l     = unique_table("l"),
    h     = unique_table("h"),
    s     = unique_table("s"), 
    sync  = unique_table("sync"), 
    Rsync = insert_table("parlooppack.RandomSync"),
    run_x = unique_table(run_name);

  /* Construct Types
     (note that type-nodes are shared!) */

  type_ptr 
    t1 = make_primitive_type(INT_T),
    t2 = make_primitive_type(VOID_T),
    t3 = make_c_i_type(make_name(NULL, Rsync, TYPE_N));

  /* Construct Declarations */
  
  expr_ptr 
   d1 = make_declexpr(give_m_t_c(0, t1, make_decl(l, 0, NULL),    LOCPAR_N)), 
   d2 = make_declexpr(give_m_t_c(0, t1, make_decl(h, 0, NULL),    LOCPAR_N)), 
   d3 = make_declexpr(give_m_t_c(0, t1, make_decl(s, 0, NULL),    LOCPAR_N)), 
   d4 = make_declexpr(give_m_t_c(0, t3, make_decl(sync, 1, NULL), LOCPAR_N)), 
   d5 = make_declexpr(give_m_t_c(0, t1, 
     make_decl(my_name -> u.name.entry, 0, make_name(NULL, l, LOCPAR_N)), LOCPAR_N)); 

  /* Construct Control */

  expr_ptr 
    args = make_op(',',  make_op(',', make_op(',', d1, d2), d3), d4),
    lw   = make_op('<',  make_name(NULL, my_name -> u.name.entry, LOCPAR_N), 
	      make_name(NULL, h, LOCPAR_N)),
    up   = make_op(PIS_, make_name(NULL, my_name -> u.name.entry, LOCPAR_N),
	      make_name(NULL, s, LOCPAR_N));

  /* Construct Method */

  stmt_ptr new_b = copy_stmt(my_loop -> u.control.s1), loop, method;

  /* Normalize Occurerence of Loop Index 
     and Handle Random Synchronization   */

  if (my_stride != 1) {
    d_hook = unique_table("ii");

    if ((new_b) && (new_b -> kind == BLOCK_S))
      new_b -> u.header.body = concat(norm_stmt(), new_b -> u.header.body); 
    else
      new_b = make_block( concat(norm_stmt(), new_b) );
  }

  my_sync = sync;

  stmt_traverse(new_b, rand_sync, (my_stride == 1) ? NULL : norm_index);

  /* Hook new method to AST */

  loop   = make_control1(FOR_S, make_op(';', d5, make_op(';', lw, up)), new_b),
  method = make_header(METHOD_S, FIN_B | (my_method -> u.header.mod & STA_B),
                          run_x, t2, 0, hook_args(args, 1, 0),
			    my_annot -> u.annot.arg, loop);
  my_class -> u.header.body = concat(my_class -> u.header.body, method);

  /* Strip Annotations from Serial Branch (after copying) */
  
  stmt_traverse(my_loop, strip_annot, NULL);

  return run_x;
}

/* STEP 3: Construction of new clas LoopWorker_x
   ********************************************* */

static void add_loopw(symb_ptr run_x) {

  /* Construct new IDs */  

  symb_ptr
   LWork    = insert_table("parlooppack.LoopWorker"),
   run      = insert_table("run"),
   poolnW   = insert_table("pool.nextWork"),
   l        = insert_table("low"),
   h        = insert_table("high"),
   s        = insert_table("stride"),
   sync     = insert_table("sync"),
   target   = insert_table("target");

  /* Construct new run() method */

  short    is_static = my_method -> u.header.mod & STA_B; 
  symb_ptr c_entry   = my_class -> u.header.name -> u.name.entry;

  expr_ptr    /* inherited fields low, high, stride, sync
		 are NOT visible in javar, hence AMBI_N */

     args = make_op(',', make_name(NULL, l,      AMBI_N),
	    make_op(',', make_name(NULL, h,      AMBI_N), 
	    make_op(',', make_name(NULL, s,      AMBI_N), 
			 make_name(NULL, sync,   AMBI_N)))),
     inv1 = make_op('i', make_name(NULL, poolnW, METH_N), make_thissup(THIS_)),
     inv2 = make_op('i',
	    make_op('.', 
	       (is_static) ? make_name(NULL, c_entry, TYPE_N) 
			   : make_name(NULL, target,  FIELD_N),
	          make_name(NULL, run_x,   MISCM_N)), hook_args(args, 0, 1));
  stmt_ptr
     mbody  = add_try( make_control1(WHILE_S, inv1, make_simple(EXP_S, inv2)) ),
     runm   = make_header(METHOD_S, PUB_B, run, make_primitive_type(VOID_T),
                          0, NULL, NULL, mbody); 

  /* Constructor */

  expr_ptr
    cargs = (is_static)  ? NULL : make_declexpr(
	    give_m_t_c(0, make_c_i_type( make_name(NULL, c_entry, TYPE_N)),
	      make_decl(target, 0, NULL), LOCPAR_N));

  stmt_ptr 
     cbody = (is_static)  ? NULL : make_simple(EXP_S, make_op('=', 
		make_op('.',  make_thissup(THIS_), 
		 make_name(NULL, target, MISCF_N)), 
		 make_name(NULL, target, LOCPAR_N)) ), 
     constr = make_header(CONS_S, 0, t_hook, NULL, 0, 
	       hook_args(cargs, 1, 0), NULL, set_args(cbody, 1));  

  /* Fields */

  stmt_ptr
    fields = (is_static)  ? NULL : 
	       give_m_t_c(0, make_c_i_type(make_name(NULL, c_entry, TYPE_N)),
	        make_decl(target, 0, NULL), FIELD_N);

  /* Link fields, constructor, and run()-method */    

  stmt_ptr classb = concat(set_args(fields, 0), concat(constr, runm));

  /* Construct new class  */

  stmt_ptr class = make_header(CLASS_S, 0, t_hook, 
		    make_c_i_type(make_name(NULL, LWork, TYPE_N)),
				     0, NULL, NULL, classb);

  /* Link Exception fields to class */

  if (f_hook)
    class -> u.header.body = concat(f_hook, class -> u.header.body);

  /* Hook new class to AST */

  comp_unit = concat(comp_unit, class);
}

/* Check Format of Loop Control
   **************************** */

static int rigid_format(expr_ptr c) {
  int rigid = 
     ( (c)                             &&
       (c -> kind == ';')              &&
       (c -> u.op.rhs)                 &&
       (c -> u.op.rhs -> kind == ';')  &&

       /* First Expression:  int VAR = LOWERBOUND */

       (c -> u.op.lhs)                             &&
       (c -> u.op.lhs -> kind == 'd')              && 
       (c -> u.op.lhs -> u.declstmt)               &&
       (c -> u.op.lhs -> u.declstmt -> u.decl.tp)  &&
       (c -> u.op.lhs -> u.declstmt -> u.decl.tp -> kind == INT_T) ); 

  if (rigid) {

    /* Get Name of Loop Index */

    my_name = c -> u.op.lhs -> u.declstmt -> u.decl.name;

    if (my_name -> u.name.info != LOCPAR_N)
      javar_out(-1, "corrupt AST in rigid_format()");

    /* Preprocessing: e1 >  e2 => e2 < e1   and
		      e1 >= e2 => e2 <= e1  conversion
       ************************************************ */

    if ( (c -> u.op.rhs -> u.op.lhs)  && 
	 ( (c -> u.op.rhs -> u.op.lhs -> kind == '>') ||
	   (c -> u.op.rhs -> u.op.lhs -> kind == GE_) ) ) {
         expr_ptr comp    = c -> u.op.rhs -> u.op.lhs;
         expr_ptr tmp     = comp -> u.op.lhs;
         comp -> kind     = 
	   (c -> u.op.rhs -> u.op.lhs -> kind == '>') ? '<' : LE_;
         comp -> u.op.lhs = comp -> u.op.rhs;
         comp -> u.op.rhs = tmp;
      }

     /* Second Expression: Ea < Eb or Ea <= Eb */

     rigid = 
       ( (c -> u.op.rhs -> u.op.lhs)              &&
         ( (c -> u.op.rhs -> u.op.lhs -> kind == '<') || 
           (c -> u.op.rhs -> u.op.lhs -> kind == LE_) ) ); 
  }
   
  if (rigid) {

     expr_ptr  stridec = c -> u.op.rhs -> u.op.rhs;

     my_stride = 0;

     /* Third Expression:  VAR++ ++VAR 
			   VAR-- --VAR
			   VAR += stride
			   VAR -= stride */

     if (stridec) {
       if ( (stridec -> kind == PP_)  || (stridec -> kind == MM_) ) {
         if  ( (is_name(stridec -> u.op.rhs, my_name))  ||
	       (is_name(stridec -> u.op.lhs, my_name))  )
           my_stride = (stridec -> kind == PP_) ? (+1) : (-1);
       }
       else if ( (stridec -> kind == PIS_) || (stridec -> kind == MINIS_) ) { 
         if  ( (is_name(stridec -> u.op.lhs, my_name))     &&
	       (stridec -> u.op.rhs)                       &&
	       (stridec -> u.op.rhs -> kind == LITERAL_)   && 
	       (stridec -> u.op.rhs -> type)               && 
	       (IS_INTEGRAL1(stridec -> u.op.rhs -> type)) && 
	       (stridec -> u.op.rhs -> u.lit.set)          &&  
	       (stridec -> u.op.rhs -> u.lit.val.l > 0)    )  
           my_stride = 
	        stridec -> u.op.rhs -> u.lit.val.l * 
	       ((stridec -> kind == PIS_) ? (+1) : (-1));
       }
     }

     rigid = my_stride != 0;
  }

  if (rigid) {
    expr_ptr comp  = c -> u.op.rhs -> u.op.lhs;

    if (my_stride > 0) {
      rigid = (is_name(comp -> u.op.lhs, my_name));
      my_bound = comp -> u.op.rhs;
    }
    else {
      rigid = (is_name(comp -> u.op.rhs, my_name));
      my_bound = c -> u.op.rhs -> u.op.lhs -> u.op.lhs;
    }

    my_init   = c -> u.op.lhs -> u.declstmt -> u.decl.init;
    my_strict = c -> u.op.rhs -> u.op.lhs -> kind == '<';
  }
  return rigid;
}

/* Actual Loop Parallelization
   *************************** */

static void parloop(void) {

  my_loop = my_annot -> u.annot.body;

  if (rigid_format(my_loop -> u.control.e)) {

       expr_ptr  mname = my_method -> u.header.name;
       char     *newcname, *newrname,
                *oldcname = id_table(my_class -> u.header.name -> u.name.entry), 
                *oldmname = (mname) ? id_table(mname -> u.name.entry) : NULL;

       if (! mname) {
	 javar_out(0, "cannot apply loop parallelization in static initializer");
	 return;
       }

#ifdef PAR_OUT
       javar_out(2, "%s-loop %s in %s()",
	    (my_annot -> u.annot.f2) ? "DOACROSS" : "DOALL", 
	    id_table(my_name -> u.name.entry), id_table(mname -> u.name.entry) );
#endif

       if ((my_annot -> u.annot.f2 > 0) && (my_stride != 1)) {
         javar_out(0, "stride-1 loop required for DOACROSS-loop");
 	 return;
       }

      /* Loop bounds must have integral type */

      if (IS_INTEGRAL1(my_init -> type)) {

        if (! (my_init -> type))
          javar_out(1, "parallelization is correct if initializer is of type int");

      }
      else {
        javar_out(0, "parallelization prohibited by type of initializer");
        return;
      }

      if (IS_INTEGRAL1(my_bound -> type)) {
        
	if (! (my_bound -> type))
          javar_out(1, "parallelization is correct if loop bound is of type int");

      }
      else {
        javar_out(0, "parallelization prohibited by type of loop bound");
        return;
      }

      /* DETACH */

      my_annot -> u.annot.body = NULL;

      /* Check Loop Invariance of Loop Bounds and Locals/Parameters */

       if ( (bounds_not_inv()) || 
	    (locpars_not_inv()) ) {
         my_annot -> u.annot.body = my_loop;
	 return;
       }

       /* STEP 1: Loop Replacement 
          STEP 2: Construction of run_x()-method 
          STEP 3: Construction of LoopWorker_x-class
	  ****************************************** */

       /* using class/method name avoids conflicts
	  between javar-generated loop workers in non-examined files
          (although not likely, conflicts with non-local programmer-defined 
	   identifiers are still possible however) */
	  
       /* DYNAMIC MEMORY ALLOCATION -> no restriction on length */

       newcname = (char *) make_mem( sizeof(char) * 
			   (strlen(oldcname) + strlen(oldmname) + 13) );
       newrname = (char *) make_mem( sizeof(char) * 
			   (strlen(oldcname) + strlen(oldmname) + 6)  );

       sprintf(newcname, "LoopWorker_%s_%s", oldcname, oldmname);
       sprintf(newrname, "run_%s_%s",        oldcname, oldmname);

       t_hook = unique_table(newcname);
       d_hook =
       p_hook = 
       s_hook = NULL;

       my_annot -> u.annot.body = replace_loop();
       add_loopw(add_run(newrname));

       /* cleanup and administration */

       free(newcname);
       free(newrname);

       par_cnt++;
   }
   else 
     javar_out(0, "parallelization prohibited by incorrect format of loop");
}

/* ***********************************************************
   *** PRIVATE FUNCTIONS RELATED TO METHOD PARALLELIZATION ***
   *********************************************************** */

/* Test Match of method-invocation with parallel method
   **************************************************** */

static short is_method(expr_ptr e) {
  if ((e) && (e -> kind == 'i')) {
    expr_ptr n = e -> u.op.lhs;

    if (n -> kind == '.') 
      n = n -> u.op.rhs;

    return ( (n -> u.name.id    == my_method -> u.header.name -> u.name.id)  &&
  	     (n -> u.name.scope == my_class  -> u.header.scope)              &&
  	     (arg_match(my_method -> u.header.e1, e -> u.op.rhs)) );
  }
  return 0;
}

/* STEP 1: Clone Parallel Method 
   ***************************** */

/* Replace all self-invocations in clone (TRAVERSAL ROUTINE) */

static void do_clone(expr_ptr e) {
  if (is_method(e)) {
    if (e -> u.op.lhs -> kind == '.') {
      e -> u.op.lhs -> u.op.rhs -> u.name.id    = 
      e -> u.op.lhs -> u.op.rhs -> u.name.entry = s_hook;
    }
    else 
      e -> u.op.lhs = make_name(e -> u.op.lhs -> u.name.lcomp, s_hook, METH_N);
  }
}

static void method_clone(void) {
  stmt_ptr clone = copy_stmt(my_method);

  stmt_traverse(clone, strip_annot, do_clone);

  clone -> u.header.name -> u.name.entry =
  clone -> u.header.name -> u.name.id    = s_hook;

  /* Hook to AST */

  my_class -> u.header.body = concat(my_class -> u.header.body, clone);
}

/* STEP 2: Construct New Driver Method 
   *********************************** */

/* Collection of Local/Parameter Declarations  */

static void collect_decl_mark(stmt_ptr s) {
  if (s -> kind == DECL_S) {
    add_list(s -> u.decl.name, s);
    mark_list(s -> u.decl.name);
  }
}

static void method_driver(void) {
 
  stmt_ptr oldb = my_method -> u.header.body, new_m;
  expr_ptr inv  = MK_LIT(0);

  /* Record Local Parameters */

  ipos = 0;
  expr_traverse(my_method -> u.header.e1, collect_decl_mark, NULL);

  /* Copy Header */

  my_method -> u.header.body = NULL;  /* DETACH */
  new_m = copy_stmt(my_method);
  my_method -> u.header.body = oldb;  /* ATTACH */

  /* Invocation */

  inv = make_op('i', make_name(NULL, p_hook, METH_N), hook_args(inv, 0, 0));

  new_m -> u.header.body = 
    make_simple((my_method -> u.header.tp -> kind != VOID_T) ? RET_S : EXP_S, inv);

  /* Hook to AST */

  my_class -> u.header.body = concat(my_class -> u.header.body, new_m);
}

/* STEP 3: Modify Original Method 
   ****************************** */

/* Replace all self-invocations in parallel method (TRAVERSAL ROUTINE) */

static void rep_all(expr_ptr e) {
  if (is_method(e)) {

    expr_ptr arg = make_op('+', MK_LIT(1), make_name(NULL, d_hook, LOCPAR_N));

    if (e -> u.op.lhs -> kind == '.') {
      e -> u.op.lhs -> u.op.rhs -> u.name.id    = 
      e -> u.op.lhs -> u.op.rhs -> u.name.entry = p_hook;
    }
    else {
      e -> u.op.lhs = make_name(e -> u.op.lhs -> u.name.lcomp, p_hook, METH_N);
    }
    e -> u.op.rhs = (e -> u.op.rhs) 
		  ? make_op(',', arg, e -> u.op.rhs)
		  : arg;
  }
}

static symb_ptr rep_dum;
static stmt_ptr pre_add, post_add;

/* Replace all occurrences of dummy (TRAVERSAL ROUTINE) */

static void rep_dummy(expr_ptr e) {
  if ((e -> kind == ID_) && (e -> u.name.info == LOCPAR_N)) {
    e -> u.name.id    =
    e -> u.name.entry = rep_dum;
  }
}

/* Make Join-construct */

static stmt_ptr join_me(expr_ptr ri, symb_ptr wi_x, symb_ptr ie_x) {
  symb_ptr 
    res = insert_table("join"),
    ire = insert_table("InterruptedException");
  expr_ptr
    dec = make_declexpr(
	    give_m_t_c(0, make_c_i_type( make_name(NULL, ire, TYPE_N)),
	      make_decl(ie_x, 0, NULL), LOCPAR_N));
  stmt_ptr 
   j = make_simple(EXP_S,
      make_op('i', 
      make_op('.', make_name(NULL, wi_x, LOCPAR_N), 
		   make_name(NULL, res, MISCM_N)), NULL) ), 
   c = make_header(CATCH_S, 0, NULL, NULL, 0, dec, NULL, NULL), 
   s = make_control(TRY_S, NULL, make_block(j), c);

  if (i_hook) {
    stmt_ptr new_i = copy_stmt(i_hook);
    rep_dum        = wi_x;
    stmt_traverse(new_i, NULL, rep_dummy);
    s              = concat(s, new_i); 
  }

  if (ri) { /* Result */

    symb_ptr res = insert_table("result");
    expr_ptr fa  = make_op('.', make_name(NULL, wi_x, LOCPAR_N),
                                make_name(NULL, res,  MISCF_N));
    s = concat(s, make_simple(EXP_S, make_op('=', ri, fa)));

    if (! s_hook) 
      add_dummy_decl(ri);
  }

  if (! s_hook) { /* No Cloning */

    expr_ptr comp = make_op(NE_, make_name(NULL, wi_x, LOCPAR_N), 
                    make_literal(-2, "null", 0));
    s = make_control(IF_S, comp, make_block(s), NULL);
  }
  return s;
}

/* Handle Level-1 Invocations  (TRAVERSAL ROUTINE) */

static void level1_meth(stmt_ptr s) {
  if (s -> kind == EXP_S) {
 
   if (--self_inv > 0) { 
   
      /* Fork-Construct to Pre-code */

      symb_ptr 
	wi_x = unique_table("wi"),
	ie_x = unique_table("ei");

      expr_ptr 
	 e     = s -> u.arg,
	 lhs   = (e -> kind == '=') ? e -> u.op.lhs : NULL,
	 mi    = (e -> kind == '=') ? e -> u.op.rhs : e,
         old_m = mi -> u.op.lhs,
         old_a = mi -> u.op.rhs,
         tar_e = (old_m -> kind == '.') 
  	       ? old_m -> u.op.lhs
	       : old_m -> u.name.lcomp,
         arg1  = make_op('+', MK_LIT(1), make_name(NULL, d_hook, LOCPAR_N)),
         arg2  = (my_method -> u.header.mod & STA_B)
	       ? arg1
	       : make_op(',', arg1, (tar_e) ? copy_expr(tar_e) : make_thissup(THIS_)),
         args  = (old_a)       
	       ? make_op(',', arg2, copy_expr(old_a))
	       : arg2,
         new   = make_te(NEW_, make_c_i_type(make_name(NULL, t_hook, TYPE_N)), args);

      stmt_ptr 
	 new_s = NULL;
  
      if (s_hook) { /* Cloning */

        new_s = give_m_t_c(0, make_c_i_type(make_name(NULL, t_hook, TYPE_N)),
	          make_decl(wi_x, 0, new), LOCPAR_N);
      }
      else {
        expr_ptr comp = make_op(LE_, make_name(NULL, d_hook, LOCPAR_N), 
                                MK_LIT(my_annot -> u.annot.f1));
  
        stmt_ptr ass  = make_simple(EXP_S, 
		        make_op('=', make_name(NULL, wi_x, LOCPAR_N), new));
  
        new_s = concat(
                  give_m_t_c(0, make_c_i_type(make_name(NULL, t_hook, TYPE_N)),
                  make_decl(wi_x, 0, make_literal(-2, "null", 0)), LOCPAR_N),
		  make_control(IF_S, comp, ass, s) );
  
      }

      pre_add  = concat(pre_add,  new_s); 
      post_add = concat(post_add, join_me(lhs, wi_x, ie_x));

    }
    else { /* Last Invocation to Pre-code */

      pre_add = concat(pre_add, s);

    }
  }
}

/* Handle Parallel Method Sequences  (TRAVERSAL ROUTINE) */

static void do_par_s(stmt_ptr s) {
  if ((s -> kind == ANNOT_S) && (s -> u.annot.f2 == -1)) {

    self_inv = s -> u.annot.f3;
    post_add = pre_add = NULL;
    
    stmt_traverse(s -> u.annot.body, level1_meth, NULL);

    /* Hook FORK/JOIN constructs to AST */

    s -> u.annot.body -> u.header.body = concat(pre_add, post_add);
  }
}

static void method_modify(void) {
  symb_ptr dumm = insert_table("dummy");
  expr_ptr decl = make_declexpr(give_m_t_c(0, make_primitive_type(INT_T),
                  make_decl(d_hook, 0, NULL), LOCPAR_N));

  /* Make Exception Lists */

  f_hook = c_hook = i_hook = NULL;

  hook_exceptions(my_annot -> u.annot.arg, dumm, NULL);

  /* Modify Body */

  stmt_traverse(my_method -> u.header.body, do_par_s, NULL);
  stmt_traverse(my_method -> u.header.body, NULL,     rep_all);

  my_method -> u.header.e1 = (my_method -> u.header.e1)
                           ? make_op(',', decl, my_method -> u.header.e1)
                           : decl;

  if (s_hook) { /* Clone */
    
    expr_ptr comp = make_op('>', make_name(NULL, d_hook, LOCPAR_N), 
	                    MK_LIT(my_annot -> u.annot.f1)),
             inv  = make_op('i', make_name(NULL, s_hook, METH_N), 
		      hook_args(NULL, 0, 0));

   stmt_ptr inv_s = make_simple((my_method -> u.header.tp -> kind != VOID_T) 
		  ? RET_S : EXP_S, inv);

    my_method -> u.header.body = 
	 make_control(IF_S, comp, inv_s, make_block(my_method -> u.header.body));
  }
  my_method -> u.header.name -> u.name.entry =
  my_method -> u.header.name -> u.name.id    = p_hook;
}

/* STEP 4: Construction of new clas TreeWorker_x
   ********************************************* */

static void method_treew(void) {

  /* Construct new IDs */  

  symb_ptr
   res = insert_table("result"),
   dep = insert_table("depth"),
   tar = insert_table("target"),
   sta = insert_table("start"),
   run = insert_table("run"),
   thr = insert_table("Thread");

  /* Construct new run() method */

  short    is_static = my_method -> u.header.mod & STA_B; 
  short    is_void   = my_method -> u.header.tp -> kind == VOID_T;
  symb_ptr c_entry   = my_class  -> u.header.name -> u.name.entry;

  expr_ptr    
     df   = make_name(NULL, dep, FIELD_N),
     inv  = make_op('i',
	    make_op('.', ((is_static) ? make_name(NULL, c_entry, TYPE_N) 
			              : make_name(NULL, tar,     FIELD_N)),
	         make_name(NULL, p_hook,   MISCM_N)), hook_args(df, 0, 1));

  stmt_ptr
     mbody  = add_try( make_simple(EXP_S, 
	      (is_void) ? inv
                        : make_op('=', make_name(NULL, res, FIELD_N), inv)) ),
     runm   = make_header(METHOD_S, PUB_B, run, make_primitive_type(VOID_T),
                          0, NULL, NULL, mbody); 

  /* Constructor */

  expr_ptr
    carg1 = make_declexpr(give_m_t_c(0, make_primitive_type(INT_T),
	    make_decl(dep, 0, NULL), LOCPAR_N)),
    carg  = (is_static)  
	  ? carg1
	  : make_op(',', carg1, 
	      make_declexpr( give_m_t_c(0, make_c_i_type( 
		make_name(NULL, c_entry, TYPE_N)),
	        make_decl(tar, 0, NULL), LOCPAR_N)));

  stmt_ptr 
    cbody1 = make_simple(EXP_S, make_op('=', 
		make_op('.',  make_thissup(THIS_), 
		 make_name(NULL, dep, MISCF_N)), 
		 make_name(NULL, dep, LOCPAR_N))), 
     cbody = (is_static)  
	   ? cbody1
	   : concat(cbody1,
	      make_simple(EXP_S, make_op('=', 
		make_op('.',  make_thissup(THIS_), 
		 make_name(NULL, tar, MISCF_N)), 
		 make_name(NULL, tar, LOCPAR_N)))), 
     constr = make_header(CONS_S, 0, t_hook, NULL, 0, 
	       hook_args(carg, 1, 0), NULL, 
	       concat(set_args( cbody, 1), 
	       make_simple(EXP_S, make_op('i', 
	       make_name(NULL, sta, METH_N), NULL))));  

  /* Fields */

  stmt_ptr
    field1 = give_m_t_c(0, make_primitive_type(INT_T),
	                   make_decl(dep, 0, NULL), FIELD_N),
    field2 = (is_void)
	   ? field1
	   : concat(field1, 
                give_m_t_c(0, copy_type(my_method -> u.header.tp),
	        make_decl(res, 0, NULL), FIELD_N)),
    fields = (is_static)  
	   ? field2
	   : concat(field2, give_m_t_c(0, 
		make_c_i_type(make_name(NULL, c_entry, TYPE_N)),
	        make_decl(tar, 0, NULL), FIELD_N));

  /* Link fields, constructor, and run()-method */    

  stmt_ptr classb = concat(set_args(fields, 0), concat(constr, runm));

  /* Construct new class  */

  stmt_ptr class = make_header(CLASS_S, 0, t_hook, 
		    make_c_i_type(make_name(NULL, thr, TYPE_N)),
				     0, NULL, NULL, classb);

  /* Link Exception fields to class */

  if (f_hook)
    class -> u.header.body = concat(f_hook, class -> u.header.body);

  /* Hook new class to AST */

  comp_unit = concat(comp_unit, class);
}

/* Test Validity of Method Parallelization (TRAVERSAL ROUTINE) */

static void cnt_annot(stmt_ptr s) {
  if (my_flag)
    return;
  else if (s -> kind == EXP_S) {
    expr_ptr arg = s -> u.arg;

    if ( (arg) && (is_method(arg)) )
      self_inv++;
    else if ( (arg) && (arg -> kind == '=') &&
	 (arg -> u.op.lhs) && (arg -> u.op.lhs -> kind == ID_)
	 		   && (arg -> u.op.lhs -> u.name.lcomp == NULL) &&
	 (arg -> u.op.rhs) && (is_method(arg -> u.op.rhs)) ) 
      self_inv++;
   else { 
     javar_out(0, "invalid statement in parallel method invocation sequence");
     my_flag = 1; 
   }
  }
  else if (s -> kind == ANNOT_S) {
    javar_out(1, "parallel annotation within sequence ignored");
    strip_annot(s);
  }
  else if ((s -> kind != CONCAT_S) && (s -> kind != BLOCK_S) &&
	   (s -> kind != EMPTY_S)) {
    javar_out(0, "invalid parallel method invocation sequence");
    my_flag = 1; 
  }
}

static void trav_m_annot(stmt_ptr s) {
  if (my_flag) 
    return;
  else if (s -> kind == ANNOT_S) {
  
    if (s -> u.annot.f2 == -1) {

      self_inv = 0;

      stmt_traverse(s -> u.annot.body, cnt_annot, NULL);

      if (my_flag)
        return;
      else if (self_inv < 2) { 
        javar_out(0, "sequence without parallel self-invocations");
        my_flag = 1;
      }
      else
        s -> u.annot.f3 = self_inv;

    }
    else {
      javar_out(1, "parallel annotation within parallel method ignored");
      strip_annot(s);
    }
  }
}

/* Actual Method Parallelization
   ***************************** */

static void parmethod(void) {
  expr_ptr  mname = my_method -> u.header.name;

  if (! mname) 
    javar_out(0, "cannot apply method parallelization to static initializer");
  else if (! my_method -> u.header.tp) 
    javar_out(0, "cannot apply method parallelization to constructor");
  else if (! (my_method -> u.header.mod & FIN_B))
    javar_out(0, "parallel method must be final");
  else if (my_method -> u.header.mod & SYN_B)
    javar_out(0, "parallel method must be un-synchronized");
  else if (my_method -> u.header.mod & PRI_B)
    javar_out(0, "parallel method must be non-private");
  else if (my_annot -> u.annot.body -> kind != BLOCK_S)
    javar_out(-1, "AST corrupt in parmethod()");
  else {

    char *newcname, *newmname,
         *oldcname = id_table(my_class -> u.header.name -> u.name.entry), 
         *oldmname = id_table(mname -> u.name.entry);

#ifdef PAR_OUT
    javar_out(2, "parallel method %s() %s", oldmname, 
		 (my_annot -> u.annot.f4) ? "[CLONE]" : "");
#endif

    my_flag = 0;

    stmt_traverse(my_method -> u.header.body, trav_m_annot, NULL);

    if (my_flag) 
      return;

    /* 
       STEP 1: Cloning of original method      -> _ser [OPTIONAL]
       STEP 2: Construction of new method      
       STEP 3: Modification of original method -> _par
       STEP 4: Construction of TreeWorker_x-class
       *********************************************************** */

    /* using class/method name avoids conflicts
       between javar-generated tree workers in non-examined files
       (although not likely, conflicts with non-local programmer-defined 
	identifiers are still possible however) */
	  
    /* DYNAMIC MEMORY ALLOCATION -> no restriction on length */

    newcname = (char *) make_mem( sizeof(char) * 
			   (strlen(oldcname) + strlen(oldmname) + 13) );
    newmname = (char *) make_mem( sizeof(char) * 
			   (strlen(oldcname) + strlen(oldmname) + 6)  );

    sprintf(newcname, "TreeWorker_%s_%s", oldcname, oldmname);
    sprintf(newmname, "PAR_%s_%s",        oldcname, oldmname);

    t_hook = unique_table(newcname);
    d_hook = unique_table("d");
    p_hook = unique_table(newmname);
    s_hook = NULL;

    if (my_annot -> u.annot.f4) { /* Clone */
      sprintf(newmname, "SER_%s_%s", oldcname, oldmname);
      s_hook = unique_table(newmname);
      method_clone();
    }

    method_driver();
    method_modify();
    method_treew();

    /* cleanup and administration */

    free(newcname);
    free(newmname);

    par_cnt++;
  }
}

/* *********************************************************
   *** PRIVATE FUNCTION  (DRIVER)                        ***
   ********************************************************* */

/* Parallelization Driver (TRAVERSAL ROUTINE)
   ****************************************** */

static void partrav(stmt_ptr s) {
  if ( (s -> kind == CLASS_S) || (s -> kind == INTERF_S) )
    my_class = s;
  else if ( (s -> kind == METHOD_S) || (s -> kind == SINIT_S) 
   			            || (s -> kind == CONS_S) ) 
    my_method = s;
  else if (s -> kind == ANNOT_S) {

    if ((! my_class) || (! my_method))
      javar_out(-1, "annotation outside any environment");
  
    my_annot = s;

    if (s -> u.annot.f2 != -1) { /* LOOP */

      if (s -> u.annot.body)
        parloop();
      else {
        javar_out(1, "%s annotation outside parallel loop ignored",
		    (s -> u.annot.arg) ? "wait" : "post");
        strip_annot(s);
      }
    }
    else { /* METHOD */

      parmethod();
    }
  }
}

/* ********************************************************
   *** PUBLIC FUNCTIONS                                 ***
   ******************************************************** */

int parallelize(void) {
  par_cnt  = 0;

  stmt_traverse(comp_unit, partrav, NULL);  

  if (par_cnt > 0) {

    /* HANDLE DECLARATIONS AND SET SCOPES RIGHT */

    del_info_table();
    semantic_validate();
  }
  return par_cnt;
}
